/*
 Standalone Temperature Logger for ATtiny85
 Version: 1.2
 Author: Alex from Inside Gadgets (http://www.insidegadgets.com)
 Created: 9/01/2011
 Last Modified: 12/02/2011
 
 Used to monitor temperature every x minutes/hours which is saved 
 to the built in EEPROM. 
 
 */

#include <avr/sleep.h>
#include <avr/wdt.h>
#include <math.h>
#include <EEPROM.h>

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

volatile boolean f_wdt = 1; // Watchdog variable

int ledPin = 0;
int thermistorPin = 1; // Analog input 1
int buttonPin = 4;
int dataPin = 1;
int clockPin = 3; // Analog input 3

int functionSelect = 0;
int dataCount = 1;
int delayTime = -1;

void setup() {
  pinMode(ledPin, OUTPUT);
  pinMode(buttonPin, INPUT);  
  pinMode(dataPin, OUTPUT);
  setup_watchdog(8);
}

void loop() {
  
  // Wait for input
  if (functionSelect == 0) {
    
    // Read delay time
    if (delayTime == -1) {
      delayTime = EEPROM.read(0);
      if (delayTime > 10) { // Not a normal value
        delayTime = 0;
      }
    }
    
    // Always reset variables in case
    digitalWrite(ledPin, LOW);
    digitalWrite(dataPin, LOW);
    
    // Check button press, if still pressed after 500ms, check again
    // after 1500ms, if still pressed go to configuration
    int buttonState = digitalRead(buttonPin);
    if (buttonState == HIGH) {
      delay(500);
      buttonState = digitalRead(buttonPin);
      if (buttonState == true) {
        delay(1500);
        buttonState = digitalRead(buttonPin);
        if (buttonState == HIGH) {
          functionSelect = 2;
          delayTime = -1;
          digitalWrite(ledPin, HIGH);
          delay(2000);
          digitalWrite(ledPin, LOW);
        }
      }
      else {
        functionSelect = 1;
        for (int x = 0; x < 3; x++) {
          digitalWrite(ledPin, HIGH);
          delay(500);
          digitalWrite(ledPin, LOW);
          delay(500);
        }
      }
    }
    
    int clockState = analogRead(clockPin);
    if (clockState >= 160) {
      int clockhighCount = 0;
      // Delay for 10 x 25ms = 250ms, check CLK is always high
      for (int x = 0; x < 10; x++) { 
        delay(25);
        clockState = analogRead(clockPin);
        if (clockState >= 160) {
          clockhighCount++;
        }
        else {
          clockhighCount = 0;
        }
      }
      // CLK is high, now give the 250ms HIGH as an ok
      if (clockhighCount >= 10) {
        functionSelect = 3;
        digitalWrite(dataPin, HIGH);
        delay(250);
        digitalWrite(dataPin, LOW);
      }
    }
  }
  
  // Log Temperature
  if (functionSelect == 1) {
    digitalWrite(dataPin, HIGH);
    int tempValue = int(Thermistor(analogRead(thermistorPin)));
    digitalWrite(dataPin, LOW);
    
    // Calculations to store -40C to 125C int to a single byte
    byte tempStore = tempValue + 130;
    if (tempValue <= -41) { // Temp is under min temp accepted
      tempStore = 90;
    }
    if (tempValue >= 126) { // Temp is over max temp accepted
      tempStore = 255;
    }
    
    // Write to EEPROM
    if (dataCount < 512) {
      EEPROM.write(dataCount, tempStore);
      dataCount++;
      if (dataCount < 512) {
        EEPROM.write(dataCount, 0); // So we know where to stop when extracting data
      }
    }
    else {
      functionSelect = 0;
    }
        
    // Calculate the delay time
    int waitCounter = 0;
    int waitTime = 0;
    f_wdt = 1;
    if (delayTime == 0) {
      waitTime = 15;
      //delay(60000);
    }
    if (delayTime >= 1 && delayTime < 4) {
      //delay((delayTime * 5) * 60000);
      waitTime = (delayTime * 5) * 15;
    }
    if (delayTime >= 4 && delayTime < 6) {
      //delay(((delayTime-3) * 30) * 60000);
      waitTime = ((delayTime-3) * 30) * 15;
    }
    if (delayTime >= 6 && delayTime < 8) {
      //delay(((delayTime-5) * 120) * 60000);
      waitTime = ((delayTime-5) * 120) * 15;
    }
    if (delayTime == 8) {
      //delay(8 * 60 * 60000);
      waitTime = 8 * 60 * 15;
    }
    if (delayTime == 9) {
      //delay(12 * 60 * 60000);
      waitTime = 12 * 60 * 15;
    }
    if (delayTime == 10) {
      //delay(24 * 60 * 60000);
      waitTime = 24 * 60 * 15;
    }
    
    while (waitCounter != waitTime) {
      if (f_wdt == 1) {
        f_wdt = 0; 
        system_sleep();
        waitCounter++;
      }
    }
  }
  
  // Configure the delay time
  if (functionSelect == 2) {
    int buttonState = digitalRead(buttonPin);
    if (buttonState == HIGH) {
      delay(500);
      int buttonPressed = false;
      buttonState = digitalRead(buttonPin);
      if (buttonState == HIGH && delayTime >= 0) {
        delay(1500);
        buttonState = digitalRead(buttonPin);
        if (buttonState == HIGH) {
          functionSelect = 0;
          EEPROM.write(0, delayTime);
          for (int x = 0; x < 3; x++) {
            digitalWrite(ledPin, HIGH);
            delay(500);
            digitalWrite(ledPin, LOW);
            delay(500);
          }
        }
        else {
          buttonPressed = true;
        }
      }
      if (buttonState == LOW || buttonPressed == true) {
        delayTime++;
        digitalWrite(ledPin, HIGH);
        delay(500);
        digitalWrite(ledPin, LOW);
        
        if (delayTime > 10) { // Too many presses
          delayTime = -1;
          digitalWrite(ledPin, HIGH);
          delay(1500);
          digitalWrite(ledPin, LOW);
        }
      }
    }
  }
  
  // Send data from EEPROM to Arduino
  if (functionSelect == 3) {
    digitalWrite(ledPin, HIGH);
    boolean inTransfer = true;
    int readCount = 1;
    while (inTransfer == true) {
      byte dataByte = 0;
      if (readCount < 512) {
        dataByte = EEPROM.read(readCount);
      }
      for (int x = 7; x >= 0; x--) {
        digitalWrite(dataPin, LOW);
        
        // Wait until CLK is low
        long resetTimer = millis();
        int clockIn = analogRead(clockPin);
        while (clockIn >= 160 && inTransfer == true) {
          clockIn = analogRead(clockPin);
          if (millis() > resetTimer + 250) {  // Check if host is restarting the CLK
            inTransfer = false;
            functionSelect = 0;
          }          
        }
        
        // Output the bit
        boolean dataBit = bitRead(dataByte, x);
        digitalWrite(dataPin, dataBit);
        
        // Wait until CLK is high
        resetTimer = millis();
        clockIn = analogRead(clockPin);
        while (clockIn == 0 && inTransfer == true) {
          clockIn = analogRead(clockPin);
          if (millis() > resetTimer + 100) {
            digitalWrite(dataPin, LOW); // Don't want to wait 250ms for this otherwise Arduino will think it's beginning from the start again
          }
          if (millis() > resetTimer + 250) { // Check if host is restarting the CLK
            inTransfer = false;
            functionSelect = 0;
          }
        }
      }
      if (dataByte == 0) { // No more data to send
        inTransfer = false;
        functionSelect = 0;
      }
      readCount++;
    }
  }
}

// Taken from http://www.arduino.cc/playground/ComponentLib/Thermistor2
double Thermistor(int RawADC) {
  double Temp;
  Temp = log(((10240000/RawADC) - 10000));
  Temp = 1 / (0.001129148 + (0.000234125 * Temp) + (0.0000000876741 * Temp * Temp * Temp));
  Temp = Temp - 273.15; // Convert Kelvin to Celcius
  return Temp;
}

// From http://interface.khm.de/index.php/lab/experiments/sleep_watchdog_battery/
void system_sleep() {
  pinMode(ledPin, INPUT);
  pinMode(dataPin, INPUT);
  
  cbi(ADCSRA,ADEN); // Switch Analog to Digital converter OFF

  set_sleep_mode(SLEEP_MODE_PWR_DOWN); // Set sleep mode
  sleep_enable(); // Go to sleep

  sleep_mode(); // System sleeps here

  sleep_disable(); // System continues execution here when watchdog timed out 
  sbi(ADCSRA,ADEN);  // Switch Analog to Digital converter ON
  
  pinMode(ledPin, OUTPUT);
  pinMode(dataPin, OUTPUT);
}

// Watchdog timeout values
// 0=16ms, 1=32ms, 2=64ms, 3=128ms, 4=250ms, 5=500ms
// 6=1sec, 7=2sec, 8=4sec, 9=8sec
// From http://interface.khm.de/index.php/lab/experiments/sleep_watchdog_battery/
void setup_watchdog(int ii) {
  byte bb;
  int ww;
  if (ii > 9 ) ii=9;
  bb=ii & 7;
  if (ii > 7) bb|= (1<<5);
  bb|= (1<<WDCE);
  ww=bb;

  MCUSR &= ~(1<<WDRF);
  // Start timed sequence
  WDTCR |= (1<<WDCE) | (1<<WDE);
  // Set new watchdog timeout value
  WDTCR = bb;
  WDTCR |= _BV(WDIE);
}
  
// Watchdog Interrupt Service / is executed when watchdog timed out
// From http://interface.khm.de/index.php/lab/experiments/sleep_watchdog_battery/
ISR(WDT_vect) {
  f_wdt = 1; // Set global flag
}
